import jax
import jax.numpy as jnp
from jax.tree_util import Partial as jax_partial

import warnings

from .pdmp import PDMP


class NonExploSpeedupZigZag(PDMP):
    """
    NonExploSpeedupZigZag class for Speedup ZigZag sampler, in the non explosive setting.
    Args:
        dim (int): The dimension of the space.
        grad_U (Callable[[Array], Array]): The gradient of the potential energy function.
        grid_size (int, optional): The number of grid points for discretizing the space. Defaults to 10.
        tmax (float, optional): The horizon for the grid. Defaults to 1.0. If 0, adaptive tmax is used.
        vectorized_bound (bool, optional): Whether to use vectorized strategy for the bound. Defaults to True.
        signed_bound (bool, optional): Whether to use signed bound strategy. Defaults to True.
        adaptive (bool, optional): Whether to use adaptive tmax Defaults to True.
        **kwargs: Additional keyword arguments.
    Attributes:
        dim (int): The dimension of the space.
        refresh_rate (float): The refresh rate.
        grad_U (Callable[[Array], Array]): The gradient of the potential.
        grid_size (int): The number of grid points for discretizing the space.
        tmax (float): The tmax for the grid.
        adaptive (bool): Whether to use adaptive tmax.
        vectorized_bound (bool): Whether to use vectorized strategy.
        signed_bound (bool): Whether to use signed strategy.
        integrator (Callable[[Array, Array, float], Tuple[Array, Array]]): The integrator function.
        rate (Array): The rate of the process.
        rate_vect (Array): The vectorized rate.
        signed_rate (Array): The signed rate.
        signed_rate_vect (Array): The vectorized and signed rate.
        velocity_jump (Callable[[Array, Array, Any], Array]): The velocity jump function.
        state (Any): The state of the ZigZag sampler.
    """
    def __init__(
        self,
        dim,
        grad_U,
        grid_size=10,
        tmax=1.0,
        vectorized_bound=True,
        signed_bound=True,
        adaptive=True,
        **kwargs,
    ):

        self.dim = dim
        self.refresh_rate = 0.0
        self.grid_size = grid_size
        self.tmax = float(tmax)
        self.vectorized_bound = vectorized_bound

        if signed_bound and (not vectorized_bound):
            self.signed_bound = False
            warnings.warn(
                "Signed bound is not compatible with non-vectorized bound for ZigZag switching to unsigned bound"
            )
        else:
            self.signed_bound = signed_bound
        self.adaptive = adaptive

        # Define the integrator for the non-explosive setting
        def integrator_path_non_explo(x, v, t):
            d = x.shape[0]
            y = x - v[0] * x[0] * v
            c = v[0] * (y @ v)
            a = (1 + y @ y) / d - (c**2) / (d**2)
            Y_0 = x[0] + (c / d)
            b_t = (Y_0 + jnp.sqrt(Y_0**2 + a)) * jnp.exp(jnp.sqrt(d) * v[0] * t)
            X_1 = (b_t**2 - a) / (2 * b_t) - (c / d)
            return y + v[0] * X_1 * v, v

        self.integrator = jax_partial(integrator_path_non_explo)
        
        # Modify the gradient of the potential used in the rate to include the change of speed
        self.speed = lambda x: jnp.sqrt(1.0 + x @ x)
        self.grad_speed = jax.grad(self.speed)
        self.true_grad_U = jax_partial(grad_U) # keep track of the true gradient
        self.grad_U = jax_partial(
            lambda x: self.speed(x) * self.true_grad_U(x) - self.grad_speed(x)
        )
        # Initialize the rate with the modified gradient
        self.rate, self.rate_vect, self.signed_rate, self.signed_rate_vect = (
            self._init_zz_rate()
        )
        # Define the velocity jump function
        def _velocity_jump_zz(x, v, key):
            lambda_t = jnp.maximum(0.0, self.grad_U(x) * v)
            proba = lambda_t / jnp.sum(lambda_t)
            m = jax.random.choice(key, jnp.arange(v.shape[0]), p=proba)
            v = v.at[m].mul(-1)
            return v

        self.velocity_jump = jax_partial(_velocity_jump_zz)

        self.state = None



import jax
import jax.numpy as jnp
from jax.tree_util import Partial as jax_partial

from typing import  Callable, Tuple
from jaxtyping import Float, Int, PRNGKeyArray, Bool, Array

from jax_tqdm import scan_tqdm

from ..namedtuples import PdmpState, PdmpOutput
from ..sampling_loop import one_step, output_state
from ..upper_bound import upper_bound_constant, upper_bound_grid, upper_bound_grid_vect

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import seaborn as sns


class PDMP:
    """
    Genereic PDMP class for sampling from a piecewise deterministic Markov process.
    Attributes:
        dim (int): The dimension of the space.
        refresh_rate (float): The refresh rate. 
        grad_U (Callable[[Array], Array]): The gradient of the potential.
        grid_size (int): The number of grid points for discretizing the space.
        tmax (float): The tmax for the grid.
        adaptive (bool): Whether to use adaptive tmax.
        vectorized_bound (bool): Whether to use vectorized strategy.
        signed_bound (bool): Whether to use signed strategy.
        integrator (Callable[[Array, Array, float], Tuple[Array, Array]]): The integrator function.
        rate (Array): The rate of the process.
        rate_vect (Array): The vectorized rate.
        signed_rate (Array): The signed rate.
        signed_rate_vect (Array): The vectorized and signed rate. 
        velocity_jump (Callable[[Array, Array, Any], Array]): The velocity jump function.
        state (Any): The state of the ZigZag sampler.
    Methods:
        init_state(xinit: Float[Array, "dim"], vinit: Float[Array, "dim"], seed: int) -> PdmpState:
        sample_skeleton(n_sk: int, xinit: Float[Array, "dim"], vinit: Float[Array, "dim"], seed: int, verbose=True) -> PdmpOutput:
        sample_from_skeleton(N: int, output: PdmpOutput) -> Float[Array, "N dim"]:
        sample(N_sk: int, N_samples: int, xinit: Float[Array, "dim"], vinit: Float[Array, "dim"], seed: int, verbose: bool = True) -> Float[Array, "N_samples dim"]:
        _init_zz_rate() -> Tuple[Callable[[Array, Array, Float], Float], Callable[[Array, Array, Float], Float], None, Callable[[Array, Array, Float], Float]]:
        _init_bps_rate() -> Tuple[Callable[[Array, Array, Float], Float], None, Callable[[Array, Array, Float], Float], None]:
    """
    def __init__(self):
        self.dim: Int
        self.refresh_rate: Float
        self.grad_U: Callable[[Array], Array]
        self.grid_size: Int
        self.tmax: Float

        self.vectorized_bound: Bool
        self.signed_bound: Bool
        self.adaptive: Bool

        self.integrator: Callable[[Array, Array, Float], Tuple[Array, Array]]
        self.rate: Callable[[Array, Array, Float], Float]
        self.velocity_jump: Callable[[Array, Array, PRNGKeyArray], Array]
        self.state: PdmpState | None

    def init_state(
        self,
        xinit: Float[Array, "dim"],
        vinit: Float[Array, "dim"],
        seed: int,
    ):
        """
        Initializes the state of the PDMP sampler.

        Args:
            xinit (Float[Array, "dim"]): The initial position.
            vinit (Float[Array, "dim"]): The initial velocity.
            seed (int): The seed for random number generation.
            upper_bound_vect (bool, optional): Whether to use vectorized upper bound function. Defaults to False.
            signed_rate (bool, optional): Whether to use signed rate function. Defaults to False.
            adaptive (bool, optional): Whether to use adaptive upper bound. Defaults to False.
            constant_bound (bool, optional): Whether to use constant upper bound. Defaults to False.

        Returns:
            PdmpState: The initialized PDMP state.
        """

        key = jax.random.key(seed)
        if self.signed_bound:
            rate = self.signed_rate
            rate_vect = self.signed_rate_vect
            refresh_rate = self.refresh_rate
        else:
            rate = self.rate
            rate_vect = self.rate_vect
            refresh_rate = 0.0
        # if the grid size is 0, we use the constant upper bound strategy using the Brent's algorithm
        if self.grid_size == 0:

            def upper_bound_func(x, v, horizon):
                func = jax_partial(lambda t: self.rate(x, v, t))
                return upper_bound_constant(func, 0.0, horizon)

        elif not self.vectorized_bound:

            def upper_bound_func(x, v, horizon):
                func = jax_partial(lambda t: rate(x, v, t))
                return upper_bound_grid(
                    func, 0.0, horizon, self.grid_size, refresh_rate
                )

        else:

            def upper_bound_func(x, v, horizon):
                func = jax_partial(lambda t: rate_vect(x, v, t))
                return upper_bound_grid_vect(func, 0.0, horizon, self.grid_size)

        upper_bound_func = jax_partial(upper_bound_func)
        boundox = upper_bound_func(xinit, vinit, self.tmax)
        state = PdmpState(
            xinit,
            vinit,
            jnp.array(0.0),
            self.tmax,
            key,  # type: ignore
            self.integrator,
            self.grad_U,
            self.rate,
            self.velocity_jump,
            upper_bound_func,
            upper_bound=boundox,
            adaptive=self.adaptive,
        )
        self.state = state
        return state

    def sample_skeleton(
        self,
        n_sk: int,
        xinit: Float[Array, "dim"],
        vinit: Float[Array, "dim"],
        seed: int,
        verbose=True,
    ) -> PdmpOutput:
        """
        Samples the skeleton of the PDMP model.

        Parameters:
        - n_sk (int): The number of skeleton samples to generate.
        - xinit (jnp.ndarray): The initial position of the particles.
        - vinit (jnp.ndarray): The initial velocity of the particles.
        - seed (int): The seed value for random number generation.
        - verbose (bool): Whether to display progress bar during sampling. Default is True.

        Returns:
        - output: The output state of the sampling process.
        """

        def one_step_inside(state, _):
            state = one_step(state)
            output = output_state(state)
            return state, output

        if verbose:
            one_step_inside = scan_tqdm(n_sk)(one_step_inside)

        initial_state = self.init_state(xinit, vinit, seed)
        initial_output = output_state(initial_state)
        state, output = jax.lax.scan(one_step_inside, initial_state, jnp.arange(n_sk))
        self.state = state
        # concatenate the initial output with the output
        output = jax.tree.map(
            lambda x, y: jnp.insert(x, 0, y, axis=0), output, initial_output
        )
        return output

    def sample_from_skeleton(self, N: int, output: PdmpOutput) -> Float[Array, "N dim"]:
        """
        Samples from the skeleton of the PDMP trajectory.

        Args:
            N (int): The number of samples to generate.
            output (PdmpOutput): The PDMP output containing the trajectory information.

        Returns:
            jnp.ndarray: The sampled points from the PDMP trajectory skeleton.
        """
        x, v, t = output.x, output.v, output.t
        t = t - t[0]
        tm = (t[-1] / N) * jnp.arange(1, N + 1)
        index = jnp.searchsorted(t, tm) - 1
        sample = jax.vmap(self.integrator)(x[index], v[index], tm - t[index])[0]
        return sample

    def sample(
        self,
        N_sk: int,
        N_samples: int,
        xinit: Float[Array, "dim"],
        vinit: Float[Array, "dim"],
        seed: int,
        verbose: bool = True,
    ) -> Float[Array, "N_samples dim"]:
        """
        Samples from the PDMP model.

        Args:
            N_sk (int): Number of skeleton points to generate.
            N_samples (int): Number of final samples to generate from the skeleton.
            xinit (jnp.ndarray): Initial position.
            vinit (jnp.ndarray): Initial velocity.
            seed (int): Seed for random number generation.
            verbose (bool, optional): Whether to print progress information. Defaults to True.

        Returns:
            jnp.ndarray: Array of samples generated from the PDMP model.
        """
        output = self.sample_skeleton(N_sk, xinit, vinit, seed, verbose)
        return self.sample_from_skeleton(N_samples, output)

    def _init_zz_rate(self):
        """
        Initializes the ZZ rate functions.

        Returns:
        - A partial function `_global_rate` that calculates the global rate given the current state.
        - A partial function `_global_rate_vect` that calculates the vectorized global rate given the current state.
        - A placeholder `_signe_rate` (currently set to None).
        - A partial function `_signed_rate_vect` that calculates the vectorized signed rate given the current state.
        """

        def _global_rate(x0, v0, t):
            xt, vt = self.integrator(x0, v0, t)
            return jnp.sum(jnp.maximum(0.0, self.grad_U(xt) * vt))

        def _global_rate_vect(x0, v0, t):
            xt, vt = self.integrator(x0, v0, t)
            return jnp.maximum(0.0, self.grad_U(xt) * vt)

        _signe_rate = None

        def _signed_rate_vect(x0, v0, t):
            xt, vt = self.integrator(x0, v0, t)
            return self.grad_U(xt) * vt

        return (
            jax_partial(_global_rate),
            jax_partial(_global_rate_vect),
            _signe_rate,
            jax_partial(_signed_rate_vect),
        )

    def _init_bps_rate(self):
        """
        Initializes the BPS rate functions.

        Returns:
            A tuple containing the following functions:
            - `_global_rate`: Computes the global rate of the BPS algorithm.
            - `_global_rate_vect`: Vectorized version of `_global_rate` (currently set to None).
            - `_signed_rate`: Computes the signed rate of the BPS algorithm.
            - `_signed_rate_vect`: Vectorized version of `_signed_rate` (currently set to None).
        """

        def _global_rate(x0, v0, t):
            xt, vt = self.integrator(x0, v0, t)
            return jnp.maximum(0.0, self.grad_U(xt) @ vt) + self.refresh_rate

        _global_rate_vect = None

        def _signed_rate(x0, v0, t):
            xt, vt = self.integrator(x0, v0, t)
            return self.grad_U(xt) @ vt + self.refresh_rate

        _signed_rate_vect = None

        return (
            jax_partial(_global_rate),
            _global_rate_vect,
            jax_partial(_signed_rate),
            _signed_rate_vect,
        )



def plot(output: PdmpOutput):
    """
    Plots various histograms based on the given PdmpOutput object. The histograms include:
    - Time between events histogram (top left)
    - Acceptance rate histogram (top right)
    - Hitting horizon histogram (bottom left)
    - Rejection histogram (bottom right)
    

    Parameters:
    - output (PdmpOutput): The PdmpOutput object containing the data to be plotted.

    Returns:
    - None
    """

    sns.set_style("ticks")
    # change font
    fig, axs = plt.subplots(2, 2, figsize=(10, 8))

    # Plot 1: Time between events histogram
    sns.histplot(np.diff(output.t), element="step", ax=axs[0, 0])
    axs[0, 0].set_title("Time between events histogram")

    # Plot 2: Acceptance rate histogram
    sns.histplot(output.ar, element="step", ax=axs[0, 1])
    axs[0, 1].set_title("Acceptance rate histogram")
    # make a vertical line for the mean and add a legend
    axs[0, 1].axvline(
        output.ar.mean(),
        color="r",
        linestyle="dashed",
        linewidth=1,
        label=f"Mean : {output.ar.mean():.3f}",
    )
    axs[0, 1].legend()

    # Plot 3: Hitting horizon histogram
    sns.histplot(
        output.hitting_horizon,
        discrete=True,
        shrink=0.8,
        stat="percent",
        element="bars",
        ax=axs[1, 0],
    )
    axs[1, 0].set_yscale("log")
    axs[1, 0].yaxis.set_major_formatter(
        ticker.FuncFormatter(lambda y, _: "{:.16g}".format(y))
    )
    axs[1, 0].set_title(
        f"Hitting horizon histogram, total : {output.hitting_horizon.sum()}"
    )

    # Plot 4: Rejection histogram
    sns.histplot(
        output.rejected,
        discrete=True,
        shrink=0.8,
        stat="percent",
        element="bars",
        ax=axs[1, 1],
    )
    axs[1, 1].set_yscale("log")
    axs[1, 1].yaxis.set_major_formatter(
        ticker.FuncFormatter(lambda y, _: "{:.16g}".format(y))
    )
    axs[1, 1].set_title(f"Rejection histogram, total : {output.rejected.sum()}")

    plt.tight_layout()
    plt.show()

    print("number of error bound : ", output.error_bound.sum())




import jax
import jax.numpy as jnp

def minimize_scalar_bounded_jax(func, bounds, xatol=1e-7, maxiter=500):
    """Jax implementation of minimize_scalar_bounded from scipy.optimize.

    Args:
        func (function): func must be able to be passed to a jax.jit function. For instance do jax.tree_util.Partial(func)
        bounds (tuple(float)): (lower_bound, upper_bound) for the minimization
        xatol (float, optional): . Defaults to 1e-5.
        maxiter (int, optional): maximum_number of iterations. Defaults to 500.
    """

    def cond_loop(val):
        (
            func,
            xatol,
            maxfun,
            sqrt_eps,
            golden_mean,
            a,
            b,
            fulc,
            nfc,
            xf,
            e,
            rat,
            fx,
            num,
            fnfc,
            ffulc,
            xm,
            tol1,
            tol2,
            golden,
        ) = val
        condi = jnp.bool_(jnp.abs(xf - xm) > (tol2 - 0.5 * (b - a)))
        return condi & (num < maxfun)

    def loop_while(val):
        (
            func,
            xatol,
            maxfun,
            sqrt_eps,
            golden_mean,
            a,
            b,
            fulc,
            nfc,
            xf,
            e,
            rat,
            fx,
            num,
            fnfc,
            ffulc,
            xm,
            tol1,
            tol2,
            golden,
        ) = val
        golden = jnp.asarray(1)

        # Check for parabolic fit
        def inner_if_1(val):
            (
                func,
                xatol,
                maxfun,
                sqrt_eps,
                golden_mean,
                a,
                b,
                fulc,
                nfc,
                xf,
                e,
                rat,
                fx,
                num,
                fnfc,
                ffulc,
                xm,
                tol1,
                tol2,
                golden,
            ) = val
            golden = 0
            r = (xf - nfc) * (fx - ffulc)
            q = (xf - fulc) * (fx - fnfc)
            p = (xf - fulc) * q - (xf - nfc) * r
            q = 2.0 * (q - r)
            p = jnp.where(q > 0.0, -p, p)
            q = jnp.abs(q)
            r = e
            e = rat
            x = xf
            # Check for acceptability of parabola
            fun_true = (
                lambda a, b, x, fulc, nfc, xf, e, rat, fx, num, fnfc, ffulc, golden: (
                    a,
                    b,
                    xf + (p + 0.0) / q,
                    fulc,
                    nfc,
                    xf,
                    e,
                    jnp.where(
                        ((x - a) < tol2) | ((b - x) < tol2),
                        tol1 * (jnp.sign(xm - xf) + ((xm - xf) == 0)),
                        (p + 0.0) / q,
                    ),
                    fx,
                    num,
                    fnfc,
                    ffulc,
                    golden,
                )
            )

            fun_false = (
                lambda a, b, x, fulc, nfc, xf, e, rat, fx, num, fnfc, ffulc, golden: (
                    a,
                    b,
                    xf,
                    fulc,
                    nfc,
                    xf,
                    e,
                    rat,
                    fx,
                    num,
                    fnfc,
                    ffulc,
                    jnp.asarray(1),
                )
            )

            a, b, x, fulc, nfc, xf, e, rat, fx, num, fnfc, ffulc, golden = jax.lax.cond(
                (jnp.abs(p) < jnp.abs(0.5 * q * r))
                & (p > q * (a - xf))
                & (p < q * (b - xf)),
                fun_true,
                fun_false,
                a,
                b,
                x,
                fulc,
                nfc,
                xf,
                e,
                rat,
                fx,
                num,
                fnfc,
                ffulc,
                golden,
            )

            return (
                func,
                xatol,
                maxfun,
                sqrt_eps,
                golden_mean,
                a,
                b,
                fulc,
                nfc,
                xf,
                e,
                rat,
                fx,
                num,
                fnfc,
                ffulc,
                xm,
                tol1,
                tol2,
                golden,
            )

        val = (
            func,
            xatol,
            maxfun,
            sqrt_eps,
            golden_mean,
            a,
            b,
            fulc,
            nfc,
            xf,
            e,
            rat,
            fx,
            num,
            fnfc,
            ffulc,
            xm,
            tol1,
            tol2,
            golden,
        )

        (
            func,
            xatol,
            maxfun,
            sqrt_eps,
            golden_mean,
            a,
            b,
            fulc,
            nfc,
            xf,
            e,
            rat,
            fx,
            num,
            fnfc,
            ffulc,
            xm,
            tol1,
            tol2,
            golden,
        ) = jax.lax.cond(jnp.abs(e) > tol1, inner_if_1, lambda _val: _val, val)

        e, rat = jax.lax.cond(
            golden == 1,
            lambda e, rat: (
                jnp.where(xf >= xm, a - xf, b - xf),
                golden_mean * jnp.where(xf >= xm, a - xf, b - xf),
            ),
            lambda e, rat: (e, rat),
            e,
            rat,
        )
        x = xf + (jnp.sign(rat) + (rat == 0)) * jnp.maximum(jnp.abs(rat), tol1)
        fu = func(x)
        num += 1

        def inner_if_2T(a, b, x, xf, fx, fulc, ffulc, nfc, fnfc):
            a = jnp.where(x >= xf, xf, a)
            b = jnp.where(x >= xf, b, xf)
            fulc, ffulc = nfc, fnfc
            nfc, fnfc = xf, fx
            xf, fx = x, fu
            return a, b, x, xf, fx, fulc, ffulc, nfc, fnfc

        def inner_if_2F(a, b, x, xf, fx, fulc, ffulc, nfc, fnfc):
            a = jnp.where(x < xf, x, a)
            b = jnp.where(x < xf, b, x)
            cond1 = (fu <= fnfc) | (nfc == xf)
            cond2 = ((fu <= ffulc) | (fulc == xf) | (fulc == nfc)) & jnp.logical_not(
                cond1
            )
            fulc = jnp.where(cond2, x, fulc)
            ffulc = jnp.where(cond2, fu, ffulc)

            fulc = jnp.where(cond1, nfc, fulc)
            ffulc = jnp.where(cond1, fnfc, ffulc)
            nfc = jnp.where(cond1, x, nfc)    
            fnfc = jnp.where(cond1, fu, fnfc)
            return a, b, x, xf, fx, fulc, ffulc, nfc, fnfc

        a, b, x, xf, fx, fulc, ffulc, nfc, fnfc = jax.lax.cond(
            fu <= fx, inner_if_2T, inner_if_2F, a, b, x, xf, fx, fulc, ffulc, nfc, fnfc
        )

        xm = 0.5 * (a + b)
        tol1 = sqrt_eps * jnp.abs(xf) + xatol / 3.0
        tol2 = 2.0 * tol1

        return (
            func,
            xatol,
            maxfun,
            sqrt_eps,
            golden_mean,
            a,
            b,
            fulc,
            nfc,
            xf,
            e,
            rat,
            fx,
            num,
            fnfc,
            ffulc,
            xm,
            tol1,
            tol2,
            golden,
        )

    xatol = jnp.asarray(xatol)
    maxfun = jnp.asarray(maxiter)
    x1, x2 = bounds
    flag = 0
    sqrt_eps = jnp.sqrt(2.2e-16)
    golden_mean = 0.5 * (3.0 - jnp.sqrt(5.0))
    a, b = x1, x2
    a = jnp.asarray(a)
    b = jnp.asarray(b)
    fulc = a + golden_mean * (b - a)
    nfc, xf = fulc, fulc
    rat = e = jnp.asarray(0.0)
    x = xf
    fx = func(x)
    num = jnp.asarray(1)

    fu = jnp.inf

    ffulc = fnfc = fx
    xm = 0.5 * (a + b)
    tol1 = sqrt_eps * jnp.abs(xf) + xatol / 3.0
    tol2 = 2.0 * tol1

    fval = fx
    golden = 0
    val = (
        func,
        xatol,
        maxfun,
        sqrt_eps,
        golden_mean,
        a,
        b,
        fulc,
        nfc,
        xf,
        e,
        rat,
        fx,
        num,
        fnfc,
        ffulc,
        xm,
        tol1,
        tol2,
        golden,
    )

    val = jax.lax.while_loop(cond_loop, loop_while, val)
    (
        func,
        xatol,
        maxfun,
        sqrt_eps,
        golden_mean,
        a,
        b,
        fulc,
        nfc,
        xf,
        e,
        rat,
        fx,
        num,
        fnfc,
        ffulc,
        xm,
        tol1,
        tol2,
        golden,
    ) = val
    fval = fx
    flag = jnp.where(num >= maxfun, False, True)
    result = {
        "min": fval,
        "argmin": xf,
        "success": flag,
        "evals": num,
        "lower_bound": a,
        "upper_bound": b,
    }
    return result



import jax
import jax.numpy as jnp
from jax.tree_util import Partial as jax_partial

from typing import  Callable
from jaxtyping import  Array


from .pdmp import PDMP


class BouncyParticle(PDMP):
    """
    BouncyParticle class for the Bouncy Particle sampler.
    Args:
        dim (int): The dimension of the space.
        grad_U (Callable[[Array], Array]): The gradient of the potential energy function.
        grid_size (int, optional): The number of grid points for discretizing the space. Defaults to 10.
        tmax (float, optional): The horizon for the grid. Defaults to 1.0. If 0, adaptive tmax is used.
        refresh_rate (float, optional): The refresh rate of the process. Defaults to 0.1.
        signed_bound (bool, optional): Whether to use signed bound strategy. Defaults to True.
        adaptive (bool, optional): Whether to use adaptive tmax Defaults to True.
        **kwargs: Additional keyword arguments.
    Attributes:
        dim (int): The dimension of the space.
        refresh_rate (float): The refresh rate.
        grad_U (Callable[[Array], Array]): The gradient of the potential.
        grid_size (int): The number of grid points for discretizing the space.
        tmax (float): The tmax for the grid.
        adaptive (bool): Whether to use adaptive tmax.
        vectorized_bound (bool): Whether to use vectorized strategy. Not used in the Bouncy Particle.
        signed_bound (bool): Whether to use signed strategy.
        integrator (Callable[[Array, Array, float], Tuple[Array, Array]]): The integrator function.
        rate (Array): The rate of the process
        rate_vect (Array): The vectorized rate. Not used in the Bouncy Particle.
        signed_rate (Array): The signed rate.
        signed_rate_vect (Array): The vectorized and signed rate. Not used in the Bouncy Particle.
        velocity_jump (Callable[[Array, Array, Any], Array]): The velocity jump function.
        state (Any): The state of the ZigZag sampler.
    """
    def __init__(
        self,
        dim: int,
        grad_U: Callable[[Array], Array],
        grid_size: int = 10,
        tmax: float = 1.0,
        refresh_rate: float = 0.1,
        signed_bound: bool = True,
        adaptive: bool = True,
        **kwargs,
    ):
        
        self.dim = dim
        self.refresh_rate = refresh_rate
        self.grad_U = jax_partial(grad_U)
        self.grid_size = grid_size
        
        # adaptive tmax if tmax is 0
        if tmax == 0:
            self.tmax = 1.0
            self.adaptive = True
        else:
            self.tmax = float(tmax)
            self.adaptive = adaptive
        self.vectorized_bound = False # not used in the Bouncy Particle
        self.signed_bound = signed_bound

        # definition of the integrator
        self.integrator = jax_partial(lambda x, v, t: (x + (v * t), v))

        # initialization of the rate
        self.rate, self.rate_vect, self.signed_rate, self.signed_rate_vect = (
            self._init_bps_rate()
        )

        # definition of the velocity jump function
        def _velocity_jump(x, v, key):
            grad_U_x = self.grad_U(x)
            dim = x.shape[0]
            bounce_prob = jnp.maximum(0.0, grad_U_x @ v)
            bounce_prob /= bounce_prob + self.refresh_rate
            subkey, subkey2 = jax.random.split(key, 2)
            u = jax.random.uniform(subkey)

            def _reflect(vect, normal):
                normal = normal / jnp.linalg.norm(normal)
                return vect - 2 * (vect @ normal) * normal

            out = jnp.where(
                u < bounce_prob,
                _reflect(v, grad_U_x),
                jax.random.normal(subkey2, shape=(dim,)),
            )
            return out

        self.velocity_jump = jax_partial(_velocity_jump)
        self.state = None




import jax
import jax.numpy as jnp
from jax.tree_util import Partial as jax_partial

from typing import  Callable
from jaxtyping import  Array


from .pdmp import PDMP


class ForwardEventChain(PDMP):
    """
    ForwardEventChain class for the Forward Event Chain sampler.
    Args:
        dim (int): The dimension of the space.
        grad_U (Callable[[Array], Array]): The gradient of the potential energy function.
        grid_size (int, optional): The number of grid points for discretizing the space. Defaults to 10.
        tmax (float, optional): The horizon for the grid. Defaults to 1.0. If 0, adaptive tmax is used.
        refresh_ortho (float, optional): The refresh rate for the orthogonal component, i.e. the probability to perform the rotation on the orthogonal space. Defaults to 0.1.
        signed_bound (bool, optional): Whether to use signed bound strategy. Defaults to True.
        adaptive (bool, optional): Whether to use adaptive tmax Defaults to True.
        **kwargs: Additional keyword arguments.
    Attributes:
        dim (int): The dimension of the space.
        refresh_rate (float): The refresh rate. Is always 0 in the Forward Event Chain.
        refresh_ortho (float): The refresh rate for the orthogonal component.
        grad_U (Callable[[Array], Array]): The gradient of the potential.
        grid_size (int): The number of grid points for discretizing the space.
        tmax (float): The tmax for the grid.
        adaptive (bool): Whether to use adaptive tmax.
        vectorized_bound (bool): Whether to use vectorized strategy.
        signed_bound (bool): Whether to use signed strategy.
        integrator (Callable[[Array, Array, float], Tuple[Array, Array]]): The integrator function.
        rate (Array): The rate of the process.
        rate_vect (Array): The vectorized rate. Not used in the Forward Event Chain.
        signed_rate (Array): The signed rate.
        signed_rate_vect (Array): The vectorized and signed rate. Not used in the Forward Event Chain.
        velocity_jump (Callable[[Array, Array, Any], Array]): The velocity jump function.
        state (Any): The state of the ZigZag sampler.
    """
    
    def __init__(
        self,
        dim,
        grad_U,
        grid_size=10,
        tmax=2.0,
        refresh_ortho=0.1,
        signed_bound=True,
        adaptive=True,
        **kwargs,
    ):  
        # Check if the dimension is greater than 2
        if dim <= 2:
            raise ValueError(
                "The dimension must be greater than 2 to use the ForwardEventChain"
            )
        self.dim = dim
        self.refresh_rate = 0.0  # no refresh rate is used in the forward event chain
        self.refresh_ortho = refresh_ortho
        self.grad_U = jax_partial(grad_U)
        self.grid_size = grid_size
        # adaptive tmax if tmax is 0
        if tmax == 0:
            self.tmax = 1.0
            self.adaptive = True
        else:
            self.tmax = float(tmax)
            self.adaptive = adaptive
        self.vectorized_bound = False # vectorized strategy is not used in the forward event chain
        self.signed_bound = signed_bound

        # define the integrator function
        self.integrator = jax_partial(lambda x, v, t: (x + (v * t), v))

        #initialize the rate
        self.rate, self.rate_vect, self.signed_rate, self.signed_rate_vect = (
            self._init_bps_rate()
        )

        # define the velocity jump function
        def _velocity_jump_event_chain(x, v, key):
            subkey1, subkey2, subkey3 = jax.random.split(key, 3)
            dim = x.shape[0]
            u = jax.random.uniform(subkey1)
            rho = -((1 - u ** (2.0 / (dim - 1.0))) ** (0.5))
            grad_U_x = self.grad_U(x)
            grad_U_x = grad_U_x / jnp.linalg.norm(grad_U_x)
            v_par = (v @ grad_U_x) * grad_U_x
            v_ortho = v - v_par

            def _refresh_ortho(key):
                g = jax.random.normal(key, shape=(2, dim))
                g1 = g[0]
                g2 = g[1]
                g1 -= (g1 @ grad_U_x) * grad_U_x
                g2 -= (g2 @ grad_U_x) * grad_U_x
                e1 = g1 / jnp.linalg.norm(g1)
                e2 = g2 - (g2 @ e1) * e1
                e2 = e2 / jnp.linalg.norm(e2)
                theta = jnp.pi / 2
                v_prop = v_ortho - (v_ortho @ e1) * e1 - (v_ortho @ e2) * e2
                v_prop += (jnp.cos(theta) * e1 + jnp.sin(theta) * e2) * (
                    e1 @ v_ortho
                ) + (jnp.sin(theta) * e1 - jnp.cos(theta) * e2) * (e2 @ v_ortho)
                v_prop /= jnp.linalg.norm(v_prop)
                v_prop *= jnp.sign(v_ortho @ v_prop)
                return v_prop

            u2 = jax.random.uniform(subkey2)
            v_prop = jnp.where(
                u2 < self.refresh_ortho,
                _refresh_ortho(subkey3),
                v_ortho / jnp.linalg.norm(v_ortho),
            )
            v_out = v_prop * (1 - rho**2) ** 0.5 + rho * grad_U_x
            return v_out

        self.velocity_jump = jax_partial(_velocity_jump_event_chain)
        self.state = None



import jax
import jax.numpy as jnp
from jax.tree_util import Partial as jax_partial


from .brent import minimize_scalar_bounded_jax
from .namedtuples import BoundBox


def upper_bound_constant(func, a, b, n_grid=100, refresh_rate=0.):
    """
    Computes the constant upper bound using the Brent's algorithm.

    Parameters:
    - func: The function for which the upper bound constant is computed.
    - a: The lower bound of the interval.
    - b: The upper bound of the interval.
    - n_grid: The number of grid points used for computation (default: 100).
    - refresh_rate: The refresh rate for the upper bound constant (default: 0).

    Returns:
    - BoundBox: An object containing the upper bound constant information.

    """
    func_min = jax_partial(lambda t: -func(t))
    bound = minimize_scalar_bounded_jax(func_min, bounds=(a, b))
    t = jnp.linspace(a, b, 2)
    box_max = jnp.atleast_1d(-bound["min"])
    box_max += refresh_rate
    cum_sum = jnp.zeros(2)
    cum_sum = cum_sum.at[1:].set(box_max * (b - a))

    return BoundBox(t, box_max, cum_sum, b - a)




def upper_bound_grid(func,a,b,n_grid = 100,refresh_rate = 0.):
    """Compute the upper bound using a grid 

    Args:
        func: the function for which the upper bound is computed
        a (float) : the lower bound of the interval
        b (float): the upper bound of the interval
        n_grid (int, optional): size of the grid for the upperbound of func. Defaults to 100.
        refresh_rate (float, optional): refresh rate for the upper bound. Defaults to 0.
    Returns:
        - BoundBox: An object containing the upper bound constant information.
    """
    t = jnp.linspace(a,b,n_grid)
    step_size = t[1]-t[0]
    func_vmap = jax.vmap(jax.value_and_grad(func))
    values,grads = func_vmap(t)
    intersection_pos = (values[:-1] - values[1:] + grads[1:]*step_size) / (grads[1:] - grads[:-1])
    intersection_pos = jnp.nan_to_num(intersection_pos,nan = 0.)
    intersection_pos = jnp.clip(intersection_pos,0.,step_size)
    intersection = values[:-1] + grads[:-1]*intersection_pos
    box_max = jnp.maximum(values[:-1],values[1:])
    box_max = jnp.maximum(box_max,intersection)
    box_max = jnp.maximum(box_max,0.)
    box_max += refresh_rate
    cum_sum = jnp.zeros(n_grid)
    cum_sum = cum_sum.at[1:].set(jnp.cumsum(box_max)*step_size)
    return BoundBox(t,box_max,cum_sum,step_size)


def upper_bound_grid_vect(func,a,b,n_grid = 100):
    """Compute the upper bound using a grid with the vectorized strategy

    Args:
        func: the function for which the upper bound is computed
        a (float) : the lower bound of the interval
        b (float): the upper bound of the interval
        n_grid (int, optional): size of the grid for the upperbound of func. Defaults to 100.
        refresh_rate (float, optional): refresh rate for the upper bound. Defaults to 0.
    Returns:
        - BoundBox: An object containing the upper bound constant information.
    """
    t = jnp.linspace(a,b,n_grid)
    step_size = t[1]-t[0]
    values,grads = jax.jvp(jax.vmap(func),(t,), (jnp.ones(t.size),))
    intersection_pos = (values[:-1] - values[1:] + grads[1:]*step_size) / (grads[1:] - grads[:-1])
    intersection_pos = jnp.nan_to_num(intersection_pos,nan = 0.)
    intersection_pos = jnp.clip(intersection_pos,0.,step_size)
    intersection = values[:-1] + grads[:-1]*intersection_pos
    box_max = jnp.maximum(values[:-1],values[1:])
    box_max = jnp.maximum(box_max,intersection)
    box_max = jnp.maximum(box_max,0.)
    cum_sum = jnp.zeros((n_grid,values.shape[1]))
    cum_sum = cum_sum.at[1:].set(jnp.cumsum(box_max,axis=0)*step_size)
    
    cum_sum = jnp.sum(cum_sum,axis = 1)
    box_max = jnp.sum(box_max,axis = 1)
    return BoundBox(t,box_max,cum_sum,step_size)



def next_event(boundbox,exp_rv):
    """
    Calculate the next event time based on the given exponential random variable and the upper bound.

    Args:
        boundbox (BoundBox): The boundbox object containing the cumulative sum and grid values.
        exp_rv (float): The exponential random variable.

    Returns:
        tuple: A tuple containing the next event time (t_prop) and the corresponding upper bound value.
    """
    
    index = jnp.searchsorted(boundbox.cum_sum,exp_rv) 
    # if the index is the last element, meaning that exp_rv > cum_sum[-1], it returns infinity
    t_prop = boundbox.grid[index-1] + (exp_rv - boundbox.cum_sum[index-1]) / (boundbox.cum_sum[index] - boundbox.cum_sum[index-1]) * boundbox.step_size
    return t_prop, boundbox.box_max[index-1]



from typing import NamedTuple, Callable, Tuple
from jaxtyping import Float, Int, PRNGKeyArray, Bool, Array
import jax.numpy as jnp


class BoundBox(NamedTuple):
    """
    Represents the output of the upper bound function. Is a NamedTuple with the following fields.

    Attributes:
        grid (Float[Array, "n_grid"]): The grid values.
        box_max (Float[Array, "n_grid - 1"]): The maximum values on each segment of the grid.
        cum_sum (Float[Array, "n_grid - 1"]): The cumulative sum of box_max.
        step_size (Float[Array, ""]): The step size of the grid.
    """
    
    grid : Float[Array, "n_grid"]
    box_max : Float[Array, "n_grid - 1"]
    cum_sum : Float[Array, "n_grid - 1"]
    step_size: Float[Array, ""]



class PdmpState(NamedTuple):
    """
    Represents the state of the PDMP sampler. Is a NamedTuple with the following fields.

    Attributes:
        x (Array[float, "dim"]): position
        v (Array[float, "dim"]): velocity
        t (Array[float, ""]): time
        horizon (Array[float, ""]): horizon
        key (PRNGKeyArray): random key
        integrator (Callable[[Array, Array, Array], Tuple[Array, Array]]): integrator function
        grad_U (Callable[[Array], Array]): gradient of the potential function
        rate (Callable[[Array, Array, Array], Array]): rate function
        velocity_jump (Callable[[Array, Array, PRNGKeyArray], Array]): velocity jump function
        upper_bound_func (Callable[[Array, Array, Array], aux.BoundBox]): upper bound function
        accept (Array[bool, ""]): accept indicator for the thinning
        upper_bound (aux.BoundBox | None): upper bound box
        indicator (Array[bool, ""]): indicator for jumping
        tp (Array[float, ""]): time to the next event
        ts (Array[float, ""]): time spent
        exp_rv (Array[float, ""]): exponential random variable for the Poisson process
        lambda_bar (Array[float, ""]): upper bound for the Poisson process
        lambda_t (Array[float, ""]): rate at the current time
        ar (Array[float, ""]): acceptance rate for the thinning
        error_bound (Array[int, ""]): count of the number of errors in the upper bound
        rejected (Array[int, ""]): count of the number of rejections in the thinning
        hitting_horizon (Array[int, ""]): count of the number of hits of the horizon
    """

    x: Float[Array, "dim"]
    v: Float[Array, "dim"]
    t: Float[Array, ""]
    horizon: Float[Array, ""]
    key: PRNGKeyArray
    integrator: Callable[[Array, Array, Array], Tuple[Array, Array]]
    grad_U: Callable[[Array], Array]
    rate: Callable[[Array, Array, Array], Array]
    velocity_jump: Callable[[Array, Array, PRNGKeyArray], Array]
    upper_bound_func: Callable[[Array, Array, Array], BoundBox]
    accept: Bool[Array, ""] = jnp.array(False)
    upper_bound: BoundBox | None = None
    indicator: Bool[Array, ""] = jnp.array(False)
    tp: Float[Array, ""] = jnp.array(0.0)
    ts: Float[Array, ""] = jnp.array(0.0)
    exp_rv: Float[Array, ""] = jnp.array(0.0)
    lambda_bar: Float[Array, ""] = jnp.array(0.0)
    lambda_t: Float[Array, ""] = jnp.array(0.0)
    ar: Float[Array, ""] = jnp.array(0.0)
    error_bound: Int[Array, ""] = jnp.array(0)
    error_value_ar: Float[Array, ""] = jnp.zeros(5)
    rejected: Int[Array, ""] = jnp.array(0)
    hitting_horizon: Int[Array, ""] = jnp.array(0)
    adaptive: Bool[Array, ""] = jnp.array(False)


class PdmpOutput(NamedTuple):
    """
    Represents the output of a PDMP (Piecewise Deterministic Markov Process) simulation. Is a NamedTuple with the following fields.

    Attributes:
        x (Float[Array, "dim"]): The state trajectory.
        v (Float[Array, "dim"]): The velocity trajectory.
        t (Float[Array, ""]): The time points at which the state and velocity are recorded.
        error_bound (Int[Array, ""]): The error bound at each time point.
        rejected (Int[Array, ""]): The indicator of whether a jump was rejected at each time point.
        hitting_horizon (Int[Array, ""]): The indicator of whether the process hit the horizon at each time point.
    """

    x: Float[Array, "dim"]
    v: Float[Array, "dim"]
    t: Float[Array, ""]
    error_bound: Int[Array, ""]
    error_value_ar: Float[Array, ""]
    rejected: Int[Array, ""]
    hitting_horizon: Int[Array, ""]
    ar: Float[Array, ""]
    horizon: Float[Array, ""]



import jax
import jax.numpy as jnp
from jax.tree_util import Partial as jax_partial

from .namedtuples import PdmpState, PdmpOutput
from .upper_bound import next_event

def output_state(state: PdmpState) -> PdmpOutput:
    """
    Converts the given PdmpState object into a PdmpOutput object by selecting the relevant fields.

    Args:
        state (PdmpState): The PdmpState object to convert.

    Returns:
        PdmpOutput: The converted PdmpOutput object.
    """
    keys = PdmpOutput._fields
    values = state._asdict()
    return PdmpOutput(**{key: values[key] for key in keys})


def compare_pdmp_states(state1, state2):
    # Convert namedtuples to dictionaries
    dict1 = state1._asdict()
    dict2 = state2._asdict()

    # Find and return differences
    for key, value in dict1.items():
        if type(value) == type(state1.integrator):
            continue
        try:
            if value != dict2[key]:
                print(f"{key} is different, value: {value} != {dict2[key]}")
        except (ValueError, TypeError):
            try:
                if jnp.any(value != dict2[key]):
                    print(f"{key} is different, value: {value} != {dict2[key]}")
            except (ValueError, TypeError):
                print(f"{key} is different, error")


def move_before_horizon(state: PdmpState) -> PdmpState:
    accept = False
    state = state._replace(accept=accept)

    def cond(state):
        return jnp.logical_and(state.tp < state.horizon, jnp.logical_not(state.accept))

    state = jax.lax.while_loop(cond, inner_while, state)
    return state


def error_acceptance(state: PdmpState) -> PdmpState:
    horizon = state.horizon / 2
    upper_bound = state.upper_bound_func(state.x, state.v, horizon)
    key, subkey = jax.random.split(state.key)
    exp_rv = jax.random.exponential(subkey)
    tp, lambda_bar = next_event(upper_bound, exp_rv)
    horizon_new = jnp.where(state.adaptive, horizon, state.horizon)
    state = state._replace(
        horizon=horizon_new,
        tp=tp,
        exp_rv=exp_rv,
        key=key,
        lambda_bar=lambda_bar,
        error_bound=state.error_bound + 1,
        error_value_ar=state.error_value_ar.at[state.error_bound % 5].set(state.ar),
    )
    return state


def ok_acceptance(state: PdmpState) -> PdmpState:
    key, subkey = jax.random.split(state.key)
    accept = jax.random.bernoulli(subkey, state.ar)
    state = state._replace(lambda_t=state.lambda_t, accept=accept, key=key)
    state = jax.lax.cond(accept, if_accept, if_not_accept, state)
    cond = jnp.logical_and(state.tp > state.horizon, jnp.logical_not(state.accept))
    state = jax.lax.cond(cond, move_to_horizon2, lambda x: x, state)
    return state


def inner_while(state: PdmpState) -> PdmpState:
    lambda_t = state.rate(state.x, state.v, state.tp)
    ar = lambda_t / state.lambda_bar
    state = state._replace(lambda_t=lambda_t, ar=ar)
    state = jax.lax.cond(ar > 1.0, error_acceptance, ok_acceptance, state)
    return state


def move_to_horizon2(state: PdmpState) -> PdmpState:
    ts = state.ts + state.horizon
    xi, vi = state.integrator(state.x, state.v, state.horizon)
    state = state._replace(x=xi, v=vi, ts=ts, hitting_horizon=state.hitting_horizon + 1)
    return state


def if_accept(state: PdmpState) -> PdmpState:
    x, v = state.integrator(state.x, state.v, state.tp)
    key, subkey = jax.random.split(state.key)
    v = state.velocity_jump(x, v, subkey)  # type: ignore
    t = state.t + state.tp + state.ts
    indicator = True
    ts = 0.0
    tp = 0.0
    accept = True
    state = state._replace(
        x=x,
        v=v,
        t=t,
        indicator=indicator,
        ts=ts,
        tp=tp,
        accept=accept,
        key=key,
    )
    return state


def if_not_accept(state: PdmpState) -> PdmpState:
    key, subkey = jax.random.split(state.key)
    exp_rv = state.exp_rv + jax.random.exponential(subkey)
    tp, lambda_bar = next_event(state.upper_bound, exp_rv)
    horizon = jnp.where(state.adaptive, state.horizon / 1.04, state.horizon)
    state = state._replace(
        tp=jnp.asarray(tp), # fix a issue when using 64 bits for some reason
        exp_rv=exp_rv,
        lambda_bar=lambda_bar,
        key=key,
        rejected=state.rejected + 1,
        horizon=horizon,
    )
    return state


def move_to_horizon(state: PdmpState) -> PdmpState:
    ts = state.ts + state.horizon
    xi, vi = state.integrator(state.x, state.v, state.horizon)
    horizon = jnp.where(state.adaptive, state.horizon * 1.01, state.horizon)
    state = state._replace(
        x=xi, v=vi, ts=ts, hitting_horizon=state.hitting_horizon + 1, horizon=horizon
    )
    return state


def one_step_while(state: PdmpState) -> PdmpState:
    upper_bound = state.upper_bound_func(state.x, state.v, state.horizon)
    key, subkey = jax.random.split(state.key)
    exp_rv = jax.random.exponential(subkey)
    tp, lambda_bar = next_event(upper_bound, exp_rv)
    cond = tp > state.horizon
    state = state._replace(
        tp=tp, exp_rv=exp_rv, lambda_bar=lambda_bar, key=key, upper_bound=upper_bound
    )
    state = jax.lax.cond(cond, move_to_horizon, move_before_horizon, state)
    return state


def one_step(state: PdmpState) -> PdmpState:
    def cond_fun(state):
        return jnp.logical_not(state.indicator)

    state = state._replace(
        error_bound=0, rejected=0, hitting_horizon=0, error_value_ar=jnp.zeros(5)
    )
    state = jax.lax.while_loop(cond_fun, one_step_while, state)
    state = state._replace(indicator=False)
    return state



from .Boomerang import Boomerang
from .BouncyParticle import BouncyParticle
from .pdmp import PDMP, plot
from .ForwardEventChain import ForwardEventChain
from .SpeedupZigZag import NonExploSpeedupZigZag
from .ZigZag import ZigZag


